from typing import List
from . import lib
import pronto

ARO = lib.get_aro_ontology()

confers_resistance_to_drug_class_rel = ARO.get_relationship('confers_resistance_to_drug_class')
confers_resistance_to_antibiotic_rel = ARO.get_relationship('confers_resistance_to_antibiotic')
has_part_rel = ARO.get_relationship('has_part')
part_of_rel = ARO.get_relationship('part_of')
regulates_rel = ARO.get_relationship('regulates')
participates_in_rel = ARO.get_relationship('participates_in')

_antibiotic_class_ids = frozenset(ar.id for ar in ARO['ARO:1000003'].subclasses(distance=1, with_self=False))
def _is_drug_class(ar: str) -> bool:
    return ar in _antibiotic_class_ids


def _get_drug_classes(super_classes_list: List[str]) -> List[pronto.Term]:
    """
    - Helper function to traverse up and record immediate child of 'antibiotic molecule' in ARO
    - Traverses up ARO until immediate child of 'antibiotic molecule' class reached and 'antibiotic mixture' class not reached
    - antibiotic molecule -> ARO:1000003
    - antibiotic mixture -> ARO:3000707
    """
    return [ar for ar in super_classes_list if ar.id in _antibiotic_class_ids and ar.id != 'ARO:3000707']


def _get_drugs(aro_num: str) -> List[pronto.Term]:
    '''
    Description: Returns a list of the drugs/antibiotics to which a gene confers resistance to.

    Parameters:
        aro_num (str): ARO number. Needs to be in the form 'ARO:number'.

    Returns:
        target (list[pronto.Term]):
            ARO nodes the drugs/antibiotics to which the input gene confers resistance to.
    '''

    target = set()

    for superclass in ARO[aro_num].superclasses():
        for drug in superclass.relationships.get(confers_resistance_to_drug_class_rel, []):
            target.add(drug)

        for drug in superclass.relationships.get(confers_resistance_to_antibiotic_rel, []):
            target.add(drug)

        for rel in [regulates_rel, participates_in_rel, part_of_rel]:
            for term in superclass.relationships.get(rel, []):
                target.update(_get_drugs(term.id))

    return sorted(target)


def confers_resistance_to(aro_num: str) -> List[str]:
    # some gene superclasses can map to drugs which are immediate children of 'antibiotic molecule'
    # only use these if they are not redundant with other drugs

    drugs = set(_get_drugs(aro_num))
    drug_classes = set(d for d in drugs if _is_drug_class(d.id))
    drugs = drugs - drug_classes

    for drug in drugs:
        cur_classes = _get_drug_classes(drug.superclasses(with_self=False))
        drug_classes -= set(cur_classes)

    drugs.update(drug_classes)

    return sorted(d.id for d in drugs)


def drugs_to_drug_classes(drugs_list: List[str]) -> List[str]:
    '''
    Description: Returns a list of categories of drug classes, e.g. cephem and penam are categorized as beta_lactam antibiotics.

    Parameters:
        drugs_list (list[str]):
            A list with ARO number of the drugs/antibiotics to which a gene confers resistance to.

    Returns:
        drug_classes (list[str]):
            A list containing the ARO numbers of the drug classes of each drug given as input to the function.
            Order of the ARO numbers of the drug classes corresponds to the order in which the drugs were given
            to the function in the drugs_list.
    '''
    drug_classes = []

    for drug in drugs_list:
        drug = ARO[drug]
        cur_drug_classes = _get_drug_classes(drug.superclasses())

        for part in drug.relationships.get(has_part_rel, []):
            cur_drug_classes += _get_drug_classes(part.superclasses())
        drug_classes.extend(set(cur_drug_classes))

    return sorted(d.id for d in drug_classes)




import pandas as pd
# Pandas <1.5 does not have copy_on_write option
try:
    pd.options.mode.copy_on_write = True
except pd.errors.OptionError:
    pass
from .drug_categorization import confers_resistance_to, drugs_to_drug_classes
from .lib import get_aro_mapping_table, get_aro_ontology
from .lib import MAPPING_TABLE_ARO_COL, TARGET_ARO_COL, DATABASES, CUT_OFF_COL
import sys
from warnings import warn


def _confers_resistance_to_or_nan(aro_id):
    import numpy as np
    r = confers_resistance_to(aro_id)
    if not r:
        return np.nan
    return r


class BaseNormalizer:
    """
    Inherit this class and customize subclass methods to implement the normalization of new databases/formats.
    """

    def __init__(self, database=None) -> None:
        self.database = database

    def run(self, input_file : str):
        """
        Main normalization pipeline.
        """
        original_annot = self.load_input(input_file)
        input_genes = self.get_input_ids(original_annot)

        aro_table = self.load_mapping_table()
        aro_table.set_index(self.preprocess_ref_genes(
            aro_table.index
        ), inplace=True)

        mapping = aro_table[MAPPING_TABLE_ARO_COL].to_dict()
        cut_offs = aro_table[CUT_OFF_COL].to_dict()

        ARO = get_aro_ontology()

        _aro_name_cache = {}
        def get_aro_name(aro_id):
            if aro_id not in _aro_name_cache:
                _aro_name_cache[aro_id] = ARO[aro_id].name
            return _aro_name_cache[aro_id]

        def aro_id_to_names(aro_ids):
            return ','.join([get_aro_name(aro_id) for aro_id in aro_ids])

        original_annot[TARGET_ARO_COL] = input_genes.map(mapping)
        original_annot['ARO_name'] = original_annot[TARGET_ARO_COL].map(
                get_aro_name,
                na_action='ignore')
        original_annot[CUT_OFF_COL] = input_genes.map(cut_offs)

        confers_resistance = original_annot[TARGET_ARO_COL].map(
                _confers_resistance_to_or_nan,
                na_action='ignore')
        resistance_to_drug_classes = confers_resistance.map(
                drugs_to_drug_classes,
                na_action='ignore')
        original_annot['confers_resistance_to'] = confers_resistance.map(
                ','.join,
                na_action='ignore')
        original_annot['confers_resistance_to_names'] = confers_resistance.map(
                aro_id_to_names,
                na_action='ignore')
        original_annot['resistance_to_drug_classes'] = resistance_to_drug_classes.map(
                ','.join,
                na_action='ignore')
        original_annot['resistance_to_drug_classes_names'] = resistance_to_drug_classes.map(
                aro_id_to_names,
                na_action='ignore')

        return original_annot

    def load_mapping_table(self):
        """
        Load the mapping table for the database, defaults to loading a single database mapping table (defined by self.database).
        """
        return get_aro_mapping_table(self.database)


    def preprocess_ref_genes(self, ref_genes):
        """
        Customize this when ref gene and input gene can not exactly match.
        """
        return ref_genes

    def load_input(self, input_file):
        """
        Customize this when it fails to parse the input data.
        """
        return pd.read_csv(input_file, sep='\t')


class ARGSOAPNormalizer(BaseNormalizer):
    def __init__(self, database=None) -> None:
        super().__init__(database='sarg')

    def get_input_ids(self, itable):
        return itable[1]

    def load_input(self, input_file):
        return pd.read_csv(input_file, sep='\t', header=None)


class DeepARGNormalizer(BaseNormalizer):
    def __init__(self, database=None) -> None:
        super().__init__(database='deeparg')

    def get_input_ids(self, itable):
        return itable['best-hit']


class ResFinderNormalizer(BaseNormalizer):
    def __init__(self, database=None) -> None:
        super().__init__(database='resfinder')

    def get_input_ids(self, itable):
        gene_identifier = 'Resistance gene'
        accession = 'Accession no.'
        return pd.Series(itable[gene_identifier] + '_' + itable[accession])

    def preprocess_ref_genes(self, ref_genes):
        split_genes = ref_genes.str.split('_')
        return pd.Series(split_genes.str[0] + '_' + split_genes.str[-1])


class AMRFinderPlusNormalizer(BaseNormalizer):
    def __init__(self, database=None) -> None:
        super().__init__(database='ncbi')

    def get_input_ids(self, itable):
        if 'Accession of closest sequence' in itable.columns and 'Sequence name' in itable.columns:
            accession = 'Accession of closest sequence'
            gene_identifier = 'Sequence name'
        elif 'Closest reference accession' in itable.columns and 'Element name' in itable.columns:
            accession = 'Closest reference accession'
            gene_identifier = 'Closest reference name'
        else:
            raise NotImplementedError(
                    'Unsupported AMRFinderPlus version detected. '
                    'Supported/tested versions are v3.10.30 or v4.0.19.')

        return pd.Series(itable[accession] + '|' + itable[gene_identifier].str.replace(' ', '_'))

    def preprocess_ref_genes(self, ref_genes):
        split_genes = ref_genes.str.split('|')
        return pd.Series(split_genes.str[1] + '|' + split_genes.str[-1])


class AbricateNormalizer(BaseNormalizer):
    def __init__(self, database=None) -> None:
        if database not in ['ncbi', 'deeparg', 'resfinder', 'megares', 'argannot']:
            raise Exception(f'{database} is not a supported database.')

        super().__init__(database)

    def get_input_ids(self, itable):
        col = dict(
            ncbi='PRODUCT',
            deeparg='best-hit',
            resfinder=dict(gene_identifier='GENE', accession='ACCESSION'),
            megares='ACCESSION',
            argannot='ACCESSION'
        )[self.database]

        if self.database == 'resfinderfg':
            return itable[col].str.split('|').str[1]
        if self.database == 'resfinder':
            return pd.Series(itable[col['gene_identifier']] + '_' + itable[col['accession']])
        if self.database == 'ncbi':
            return pd.Series(itable[col].str.replace(' ', '_'))
        return itable[col]

    def preprocess_argannot_ref_genes(self, ref_gene):
        split_str = ref_gene.split(':')
        if not str(split_str[2][0]).isnumeric() and '-' not in split_str[2]:
            return ':'.join([split_str[1], split_str[3]])
        return ':'.join(ref_gene.split(':')[1:3])

    def preprocess_ref_genes(self, ref_genes):
        process_funcs_by_db = dict(
            ncbi=lambda x: x.split('|')[-1],
            deeparg=lambda x: x,
            resfinder=lambda x: '_'.join([x.split('_')[0], x.split('_')[1], x.split('_')[-1]]),
            sarg=lambda x: x,
            megares=lambda x: x.split('|')[0],
            argannot=self.preprocess_argannot_ref_genes,
            resfinderfg=lambda x: x.split('|')[1]
        )
        return ref_genes.map(process_funcs_by_db[self.database])


class GrootNormalizer(BaseNormalizer):
    def __init__(self, database=None) -> None:
        if database not in ['groot-argannot', 'groot-resfinder', 'groot-db', 'groot-core-db', 'groot-card']:
            raise NotImplementedError(f'{database} is not a supported database for groot.')
        super().__init__(database)

    def load_input(self, input_file):
        return pd.read_csv(input_file, sep='\t', header=None)

    def preprocess_groot_db_inputs(self, gene_name):
        processed_gene_name = str(gene_name).split('__')[1]
        if 'card' in  str(gene_name).split('__')[0].lower():
            processed_gene_name = processed_gene_name.split('|')[-1]
        return processed_gene_name

    def get_input_ids(self, itable):
        if self.database == 'groot-argannot':
            return itable[0].map(lambda x: x.split('~~~')[-1])

        if self.database == 'groot-card':
            return itable[0].map(lambda x: x.split('.')[0])

        if self.database in ['groot-db', 'groot-core-db']:
            return itable[0].map(self.preprocess_groot_db_inputs)

        return itable[0]

    def preprocess_ref_genes(self, ref_genes):
        if self.database == 'groot-argannot':
            return ref_genes.map(lambda x: ':'.join(str(x).split(':')[1:3]))
        return ref_genes


class HamronizationNormalizer(BaseNormalizer):
    def __init__(self, database=None, skip_on_unsupported_tool=False):
        super().__init__(database)
        self.skip_on_unsupported_tool = skip_on_unsupported_tool

    def preprocess_groot_db_inputs(self, gene_name):
        processed_gene_name = str(gene_name).split('__')[1]
        if 'card' in  str(gene_name).split('__')[0].lower():
            processed_gene_name = processed_gene_name.split('|')[-1]
        return processed_gene_name

    def preprocess_argannot_ref_genes(self, ref_gene):
        split_str = ref_gene.split(':')
        if not str(split_str[2][0]).isnumeric() and not '-' in split_str[2]:
            return ':'.join([split_str[1], split_str[3]])
        return ':'.join(ref_gene.split(':')[1:3])

    def get_input_ids(self, original_annot):
        input_id_lookup = {
            'argsoap': lambda x: x['reference_accession'],
            'deeparg': lambda x: x['gene_name'],
            'resfinder': lambda x: x['gene_name'] + '_' + x['reference_accession'],
            'amrfinderplus': lambda x: str(x['gene_name']).replace(' ', '_'),
            'groot': {
                'groot-card': lambda x: x['gene_name'].split('.')[0],
                'groot-argannot': lambda x: x['gene_name'].split('~~~')[-1],
                'groot-db': self.preprocess_groot_db_inputs,
                'groot-resfinder': lambda x: x['gene_name']
            },
            'abricate': {
                'sarg': lambda x: x['gene_symbol'],
                'megares': lambda x: x['reference_accession'],
                'argannot': lambda x: x['reference_accession'],
                'resfinderfg': lambda x: x['gene_name'].split('|')[1],
                'deeparg': lambda x: x['gene_name'],
                'resfinder': lambda x: x['gene_name'] + '_' + x['reference_accession'],
                'ncbi': lambda x: str(x['gene_name']).replace(' ', '_')
            }
        }
        input_genes = []
        for _,row in original_annot.iterrows():
            analysis_software = row['analysis_software_name'].lower() \
                                    .replace('-', '') \
                                    .replace('_', '') \
                                    .replace(' ', '')

            if 'ncbi' in analysis_software:
                analysis_software = 'amrfinderplus'

            for tool in input_id_lookup:
                if tool in analysis_software:
                    analysis_software = tool
                    break
            else:
                if self.skip_on_unsupported_tool:
                    warn(f'{tool} is not a supported ARG annotation tool. Skipping this row.')
                    input_genes.append(None)
                    continue
                else:
                    sys.stderr.write(f'{analysis_software} is not a supported ARG annotation tool\n')
                    sys.stderr.write(f'argNorm can only map genes from the following tools: {list(input_id_lookup.keys())}\n')
                    sys.exit(1)

            if analysis_software == 'groot':
                if '~~~' in row['gene_name']:
                    input_genes.append(input_id_lookup[analysis_software]['groot-argannot'](row))
                elif '.' in row['gene_name']:
                    input_genes.append(input_id_lookup[analysis_software]['groot-card'](row))
                elif 'groot-db_' in row['gene_name']:
                    input_genes.append(input_id_lookup[analysis_software]['groot-db'](row['gene_name']))
                else:
                    input_genes.append(input_id_lookup[analysis_software]['groot-resfinder'](row))
            elif analysis_software == 'abricate':
                try:
                    database = row['reference_database_id']
                except KeyError:
                    try:
                        database = row['reference_database_name']
                    except KeyError:
                        sys.stderr.write(f'An unrecognized hamronization format has been detected. Please use hamronization v1.1.8 or v1.0.4')
                        sys.exit(1)
                input_genes.append(input_id_lookup[analysis_software][database](row))
            else:
                input_genes.append(input_id_lookup[analysis_software](row))
        return pd.Series(input_genes)

    def load_mapping_table(self):
        preprocess_ref_genes_funcs = {
            'resfinder': lambda x: x.split('_')[0] + '_' + x.split('_')[-1],
            'ncbi': lambda x: x.split('|')[-1],
            'groot': lambda x: ':'.join(str(x).split(':')[1:3]) if ':' in x else x,
            'deeparg': lambda x: x,
            'sarg': lambda x: x,
            'megares': lambda x: x.split('|')[0],
            'argannot': self.preprocess_argannot_ref_genes,
            'resfinderfg': lambda x: x.split('|')[1],
        }

        mapping_tables = []
        for db in DATABASES:
            table = get_aro_mapping_table(db)

            # walrus operator was added in 3.8, so we can use it once we drop
            # support for 3.7
            ref_gene_f = preprocess_ref_genes_funcs.get(db)
            if ref_gene_f is not None:
                table.index = table.index.map(ref_gene_f)
            mapping_tables.append(table)

        return pd.concat(mapping_tables)



import argparse
import sys
import textwrap

from .argnorm_version import __version__, is_release
from .atomicwrite import atomic_write
from . import lib


def main():
    """
    Major function to run when running `argnorm` in shell.
    """
    parser = argparse.ArgumentParser(
        description=('argNorm normalizes ARG annotation results from '
                     'different tools and databases to the same ontology, '
                     'namely ARO (Antibiotic Resistance Ontology).'),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent('''\

        If argNorm is helpful in a scientific publication, please cite:

            argNorm: normalization of antibiotic resistance gene annotations to the Antibiotic Resistance Ontology (ARO)
            by Svetlana Ugarcina Perovic, Vedanth Ramji, Hui Chong, Yiqian Duan, Finlay Maguire, Luis Pedro Coelho
            in Bioinformatics (2025) https://doi.org/10.1093/bioinformatics/btaf173
         '''))

    parser.add_argument('--version', '-v',
                        action='version',
                        version=(__version__ if is_release else f'{__version__}-dev')
                        )
    parser.add_argument('tool', type=str.lower,
                        choices=['argsoap', 'abricate', 'deeparg', 'resfinder', 'amrfinderplus', 'groot', 'hamronization'],
                        help='tool (required): The bioinformatics tool used for ARG annotation.')
    parser.add_argument('--db', type=str.lower,
                        choices=lib.DATABASES,
                        help='--db (mostly optional): The database used alongside the ARG annotation tool. This is\
                            only required if abricate or groot is used as a tool. Please refer here for more information\
                                on --db: https://github.com/BigDataBiology/argNorm?tab=readme-ov-file#--db-optional')
    parser.add_argument('-i', '--input', type=str,
                        help='-i (required): The path to the ARG annotation result which needs to be normalized.')
    parser.add_argument('--hamronization_skip_unsupported_tool', action='store_true',
                        help="--hamronization_skip_unsupported_tool (optional): skip rows with unsupported tools\
                            for hamronization outputs. argNorm be default will raise an exception if unsupported\
                             tool is found in hamronization. Use this if you only want argNorm to raise a warning.")
    parser.add_argument('-o', '--output', type=str,
                        help="-o (required): The path to the output file where you would like to store argNorm's results")
    args = parser.parse_args()

    if args.output is None:
        sys.stderr.write('Please specify an output file using `-o` or `--output`\n')
        sys.exit(2)

    if args.tool in ['groot', 'abricate'] and args.db == None:
        sys.stderr.write('Please specify a database using `--db` when using groot or abricate\n')
        sys.exit(2)

    # We only import the normalize function when the user actually wants to run the program
    # This makes running `argnorm -h` much faster because it avoids importing slow modules (e.g. pandas)
    from .normalize import normalize
    result = normalize(args.input,
            tool=args.tool,
            database=args.db,
            skip_on_unsupported_tool=args.hamronization_skip_unsupported_tool
        )

    prop_unmapped = result.ARO.isna().mean()
    print(f'{args.output}:', f'{1 - prop_unmapped:.1%} ARGs mapped.')
    with atomic_write(args.output, mode='w', overwrite=True) as out:
        out.write(f'# argNorm version: {__version__}\n')
        result.to_csv(out, sep='\t', index=False)



import pytest
from argnorm import lib
from argnorm import drug_categorization

ARO = lib.get_aro_ontology()

@pytest.mark.parametrize('database', lib.DATABASES)
def test_database(database):
    determinant_of_ab_resistance = ARO['ARO:3000000']
    antibiotic_molecule = ARO['ARO:1000003']
    process_or_component_of_antibiotic_biology_or_chemistry = ARO['ARO:1000001']
    assert determinant_of_ab_resistance.name == 'determinant of antibiotic resistance'
    assert antibiotic_molecule.name == 'antibiotic molecule'
    assert process_or_component_of_antibiotic_biology_or_chemistry.name == "process or component of antibiotic biology or chemistry"

    db = lib.get_aro_mapping_table(database)
    assert len(db.index) == len(set(db.index)), f'Duplicate gene names in {database}'

    for ar in set(db['ARO'].dropna()):
        assert ar in ARO, f'ARO not found in the ontology: {ar}'

        assert (
                determinant_of_ab_resistance in ARO[ar].superclasses()
                or
                process_or_component_of_antibiotic_biology_or_chemistry in ARO[ar].superclasses()
                )

        for d in drug_categorization.confers_resistance_to(ar):
            assert d in ARO, f'ARO not found in the ontology: {d}'
            assert antibiotic_molecule in ARO[d].superclasses(), f'ARO not an antibiotic molecule: {d}'
            for dc in drug_categorization.drugs_to_drug_classes([d]):
                assert dc in ARO, f'ARO not found in the ontology: {dc}'
                assert antibiotic_molecule in  ARO[dc].superclasses(1), f'ARO term not an immediate child of antibiotic molecule: {dc}'

def test_misannotated_genes():
    assert lib.map_to_aro('(Tet)tetH:EF460464:6286-7839:1554', 'argannot') == ARO['ARO:3000175']



from .normalizers import ARGSOAPNormalizer, \
    DeepARGNormalizer, AbricateNormalizer, ResFinderNormalizer, AMRFinderPlusNormalizer, GrootNormalizer, HamronizationNormalizer

def normalize(ifile, tool : str, database : str, skip_on_unsupported_tool : bool):
    '''Normalize ARG tables

    Parameters
    ----------

    ifile : input file
    tool : str
    database : str
        Database to use
    is_hamronized : bool
        Whether input has been run through hAMRonization already
    '''
    normalizer = {
        'abricate': AbricateNormalizer,
        'amrfinderplus': AMRFinderPlusNormalizer,
        'argsoap': ARGSOAPNormalizer,
        'deeparg': DeepARGNormalizer,
        'resfinder': ResFinderNormalizer,
        'groot': GrootNormalizer,
        'hamronization': HamronizationNormalizer
    }.get(tool)

    if normalizer is None:
        raise ValueError('Please specify a correct tool name.')
    
    if normalizer == HamronizationNormalizer:
        norm = normalizer(database=database, skip_on_unsupported_tool=skip_on_unsupported_tool)
    else:
        norm = normalizer(database=database)
    
    return norm.run(ifile)




import os
# do not import pandas here because it is slow to import

# import here so that __version__ is available in the package
from .argnorm_version import __version__

ORIGINAL_ID_COL = 'Original ID'
MAPPING_TABLE_ARO_COL = 'ARO'
TARGET_ARO_COL = 'ARO'
CUT_OFF_COL = "Cut_Off"

DATABASES = [
    'argannot',
    'deeparg',
    'megares',
    'ncbi',
    'resfinder',
    'resfinderfg',
    'sarg',
    'groot-db',
    'groot-core-db',
    'groot-argannot',
    'groot-resfinder',
    'groot-card',
]

_ROOT = os.path.abspath(os.path.dirname(__file__))
_ARO = None

def get_aro_ontology():
    """
    The ARO ontology used by (bundled with) argNorm

    Returns:
        ARO (pronto.Ontology): A pronto ontology object with ARO terms.
    """
    import pronto
    import importlib.resources
    global _ARO
    if _ARO is None:
        _ARO = pronto.Ontology(os.path.join(_ROOT, 'data/aro.obo'))
    return _ARO


def get_aro_mapping_table(database):
    """
    Description: Returns the ARO mapping table for a specific supported databases.

    Parameters:
        database (str): name of database. Can be: argannot, deeparg, megares, ncbi, resfinderfg and sarg

    Returns:
        aro_mapping_table (DataFrame): A pandas dataframe with ARGs mapped to AROs.
    """
    import pandas as pd

    if 'groot' in database:
        database = 'groot'

    aro_mapping_table = pd.read_csv(
            os.path.join(_ROOT, 'data', f'{database}_ARO_mapping.tsv'),
            sep='\t', dtype={'ARO': str})
    aro_mapping_table.drop_duplicates(subset=['Original ID'], inplace=True)
    aro_mapping_table.set_index('Original ID', inplace=True)

    manual_curation = pd.read_csv(
                    os.path.join(_ROOT, 'data/manual_curation', f'{database}_curation.tsv'),
                    sep='\t', index_col=0, dtype={'ARO': str})
    if database != 'groot':
        manual_curation['Database'] = aro_mapping_table['Database'].iloc[0]
    manual_curation[CUT_OFF_COL] = 'Manual'
    aro_mapping_table.drop(index=set(manual_curation.index) & set(aro_mapping_table.index), inplace=True)
    aro_mapping_table = pd.concat([aro_mapping_table, manual_curation])

    aro_mapping_table['ARO'] = aro_mapping_table['ARO'].map(lambda a: f'ARO:{a}', na_action='ignore')
    return aro_mapping_table

def map_to_aro(gene, database):
    """
    Description: Gets ARO mapping for a specific gene in a database.

    Parameters:
        gene (str): The original ID of the gene as mentioned in source database.
        database (str): name of database. Can be: argannot, deeparg, megares, ncbi, resfinderfg, sarg, groot-db, groot-core-db, groot-argannot, groot-resfinder, groot-card

    Returns:
        ARO[result] (pronto.term.Term): A pronto term with the ARO number of input gene. ARO number can be accessed using 'id' attribute and gene name can be accessed using 'name' attribute.

        If ARO mapping is doesn't exist, None is returned.
    """
    import pandas as pd

    if database not in DATABASES:
        raise Exception(f'{database} is not a supported database.')

    mapping_table = get_aro_mapping_table(database)

    # Preprocess input gene & mapping table original ids if groot is being used
    if database == 'groot-argannot':
        gene = gene.split('~~~')[-1]
        mapping_table.index = mapping_table.index.map(lambda x: ':'.join(str(x).split(':')[1:3]))
    if database == 'groot-card':
        gene = gene.split('.')[0]
    if database in ['groot-db', 'groot-core-db']:
        if 'card' in gene.lower():
            gene = gene.split('|')[-1]
        else:
            gene = gene.split('__')[1]

    try:
        result = mapping_table.loc[gene, 'ARO']
    except KeyError:
        raise Exception(f'{gene} is not in {database} database')
    else:
        if pd.isna(result):
            return None
        ARO = get_aro_ontology()
        return ARO.get(result)



import pytest
import argnorm.normalize as argnorm
import pandas as pd
import os

def get_normed(normalizer, input_path):
    normed = normalizer.run(input_path)
    assert not (normed['ARO'] == 'ARO:nan').any()
    return normed


def get_golden_file(category, file_name):
    path = os.path.join('./outputs/', category, file_name)
    return pd.read_csv(path, sep='\t', skiprows=1)

def test_argsoap_normalizer():
    normalizer = argnorm.ARGSOAPNormalizer()
    input_path = './examples/raw/args-oap.sarg.reads.tsv'

    normed = get_normed(normalizer, input_path)
    golden_file = get_golden_file('raw', 'args-oap.sarg.reads.tsv')
    # The numbers in the heading of the raw normed df have the type int, while numbers in heading of golden file have type str
    normed.columns = golden_file.columns

    pd.testing.assert_frame_equal(normed, golden_file)

def test_deeparg_normalizer():
    normalizer = argnorm.DeepARGNormalizer()
    input_path = './examples/raw/deeparg.deeparg.orfs.tsv'

    normed = get_normed(normalizer, input_path)
    golden_file = get_golden_file('raw', 'deeparg.deeparg.orfs.tsv')

    pd.testing.assert_frame_equal(normed, golden_file)


@pytest.mark.parametrize('database', ['argannot', 'megares', 'ncbi', 'resfinder'])
def test_abricate_normalizer(database):
    normalizer = argnorm.AbricateNormalizer(database=database)
    input_path = f'./examples/raw/abricate.{database}.tsv'

    normed = get_normed(normalizer, input_path)
    golden_file = get_golden_file('raw', f'abricate.{database}.tsv')

    pd.testing.assert_frame_equal(normed, golden_file)


@pytest.mark.parametrize('database', ['random_db', 'sarg', 'resfinderfg'])
def test_abricate_validation(database):
    with pytest.raises(Exception):
        argnorm.AbricateNormalizer(database=database)


@pytest.mark.parametrize('mode', ['reads', 'orfs'])
def test_resfinder_normalizer(mode):
    normalizer = argnorm.ResFinderNormalizer()
    input_path = f'./examples/raw/resfinder.resfinder.{mode}.tsv'
    
    normed = get_normed(normalizer, input_path)
    golden_file = get_golden_file('raw', f'resfinder.resfinder.{mode}.tsv')

    pd.testing.assert_frame_equal(normed, golden_file)


@pytest.mark.parametrize('version', ['v3.10.30', 'v4'])
def test_amrfinderplus_normalizer(version):
    normalizer = argnorm.AMRFinderPlusNormalizer()
    input_file = f'./examples/raw/amrfinderplus.ncbi.orfs.{version}.tsv'

    normed = get_normed(normalizer, input_file)
    golden_file = get_golden_file('raw', f'amrfinderplus.ncbi.orfs.{version}.tsv')

    pd.testing.assert_frame_equal(normed, golden_file)


@pytest.mark.parametrize('database', ['argannot', 'resfinder', 'card', 'groot-db', 'groot-core-db'])
def test_groot_normalizer(database):
    if 'groot' not in database:
        groot_db = f'groot-{database}'
    else:
        groot_db = database

    normalizer = argnorm.GrootNormalizer(database=groot_db)
    input_path = f'./examples/raw/groot.{database}.tsv'
    normed = get_normed(normalizer, input_path)
    normed.columns = normed.columns.astype(str)
    golden_file = get_golden_file('raw', f'groot.{database}.tsv')
    pd.testing.assert_frame_equal(normed, golden_file)


def test_hamronization_normalizer():
    normalizer = argnorm.HamronizationNormalizer()

    for file in os.listdir('./examples/hamronized/'):
        if 'abricate.card' in file or 'args-oap.sarg.orfs' in file:
            continue
        if file == 'combined_hamronization_full.tsv':
            normalizer = argnorm.HamronizationNormalizer(skip_on_unsupported_tool=True)

        normed = get_normed(normalizer, f'./examples/hamronized/{file}')
        golden_file = get_golden_file('hamronized', file)
        pd.testing.assert_frame_equal(normed, golden_file)



import pytest
from argnorm import lib
from argnorm.lib import map_to_aro, get_aro_mapping_table
import pandas as pd

ARO = lib.get_aro_ontology()

def test_map_to_aro():
    test_cases = [
        ["(AGly)AAC(6')-Isa:NG_047311:101-574:474", 'argannot'],
        ["MEG_21|Drugs|Aminoglycosides|Aminoglycoside_N-acetyltransferases|AAC3", 'megares'],
        ["1028085756|WP_063844287.1|1|1|cpt|cpt|phosphotransferase|2|CHLORAMPHENICOL|PHENICOL|chloramphenicol_phosphotransferase_CPT", 'ncbi'],
        ["gb|AAG57600.1|ARO:3000318|mphB", "sarg"],
        ["(Phe)cpt_strepv:U09991:AAB36569:1412-1948:537", "argannot"],
        ["MEG_4060|Metals|Multi-metal_resistance|Multi-metal_resistance_protein|MREA", "megares"],
        ["gi:447201629:ref:WP_001278885.1:|FEATURES|cob(I)alamin_adenolsyltransferase|unclassified|cob(I)alamin_adenolsyltransferase", "deeparg"],
        ["argannot~~~(Bla)cfxA4~~~AY769933:1-966", 'groot-argannot'],
        ["ErmF.3000498.M17124.1181-1982.593", 'groot-card'],
        ["groot-db_RESFINDER__tet(W)_1_DQ060146", 'groot-db']
    ]

    expected_output = [
        ARO.get_term('ARO:3002563'),
        ARO.get_term('ARO:3004623'),
        ARO.get_term('ARO:3000249'),
        ARO.get_term('ARO:3000318'),
        ARO.get_term('ARO:3000249'),
        None,
        ARO.get_term('ARO:0010004'),
        ARO.get_term('ARO:3003005'),
        ARO.get_term('ARO:3000498'),
        ARO.get_term('ARO:3000194')
    ]

    for t, e in zip(test_cases, expected_output):
        if t[1] == 'groot':
            assert map_to_aro(t[0], t[1], t[2]) == e
        else:
            assert map_to_aro(t[0], t[1]) == e

@pytest.mark.parametrize('database', ['argannot', 'megares', 'ncbi', 'resfinder', 'resfinderfg', 'groot', 'groot-argannot'])
def test_get_aro_mapping_table_smoke(database):
    df = get_aro_mapping_table(database)
    assert len(df) > 0

def test_megares_mappings():
    """
    Megares annotations are derived from argannot and resfinder mappings (see db_hamronisation/construct_megares_mappings.py).
    Checking to see if megares mappings taken directly from argannot and resfinder are correct/updated because they can be missed
    when running `construct_megares_mappings.py` as it depends on existing argnorm mappings.
    """
    
    # "megares_resfinder_argannot_mapping.tsv" is created in "db_harmonisation/construct_megares_mappings.py"
    # It can be found in "db_harmonisation/mapping/" directory which is created when "crude_db_harmonisation.py" is run
    megares_mappings = pd.read_csv('./tests/megares_mappings/megares_resfinder_argannot_mapping.tsv', sep='\t').dropna()
    mapped_genes = list(megares_mappings['Original ID'].map(lambda x: map_to_aro(x, 'megares')))
    mapped_aros = list(megares_mappings['ARO'].map(lambda x: ARO.get_term(f'ARO:{int(x)}')))    
    assert mapped_genes == mapped_aros



from argnorm.drug_categorization import (
    confers_resistance_to,
    drugs_to_drug_classes
)
from argnorm.lib import get_aro_ontology

def _assert_aro_name(aro_id, name):
    ARO = get_aro_ontology()
    assert ARO[aro_id].name == name
    return aro_id

def test_confers_resistance_to():
    test_cases = [
        "ARO:3003938",
        "ARO:3000014",
        "ARO:3003725",
        "ARO:3000230",
        "ARO:3005046",
        "ARO:3001414",
        "ARO:3003548",
        "ARO:3000826",
        "ARO:3003066",
        "ARO:3003745"
    ]

    expected_output = [
        ["ARO:3000008", "ARO:0000032"],
        ["ARO:3000008", "ARO:0000004", "ARO:0000032", "ARO:3000637"],
        ['ARO:0000028', 'ARO:0000029'],
        ["ARO:3007382", "ARO:0000052", "ARO:0000007", "ARO:0000035", "ARO:0000049"],
        ["ARO:0000015", "ARO:3000008"],
        ['ARO:0000032', 'ARO:0000056', 'ARO:3000008'],
        ['ARO:0000045', 'ARO:0000047'],
        ['ARO:0000001', 'ARO:0000030', 'ARO:0000051', 'ARO:3000169', 'ARO:3000385', 'ARO:3000637', 'ARO:3000704', 'ARO:3000870'],
        ['ARO:0000016', 'ARO:0000032', 'ARO:3000008'],
        ['ARO:0000000', 'ARO:3000668']
    ]

    for t,e in zip(test_cases, expected_output):
        assert sorted(confers_resistance_to(t)) == sorted(e)

def test_oxa19_drugs():
    oxa_19 = _assert_aro_name('ARO:3001414', 'OXA-19')
    cephalosporin = _assert_aro_name('ARO:0000032', 'cephalosporin')
    penam = _assert_aro_name('ARO:3000008', 'penicillin beta-lactam')
    oxacillin = _assert_aro_name('ARO:0000056', 'oxacillin')
    assert sorted(confers_resistance_to(oxa_19)) == sorted([cephalosporin, penam, oxacillin])

def test_aadb_drugs():
    # aadb is a synonym for ANT(2'')-Ia
    aadb = _assert_aro_name('ARO:3000230', "ANT(2'')-Ia")
    gentamicin = _assert_aro_name('ARO:3007382', 'gentamicin')
    tobramycin = _assert_aro_name('ARO:0000052', 'tobramycin')
    dibekacin = _assert_aro_name('ARO:0000007', 'dibekacin')
    sisomicin = _assert_aro_name('ARO:0000035', 'sisomicin')
    kanamycin_a = _assert_aro_name('ARO:0000049', 'kanamycin A')
    assert sorted(confers_resistance_to(aadb)) == sorted([
        gentamicin,
        tobramycin,
        dibekacin,
        sisomicin,
        kanamycin_a
    ])

def test_drug_to_drug_classes():
    test_cases = [
        ["ARO:0000004"],
        ["ARO:3000157"],
        ["ARO:0000030"],
        ["ARO:0000036"],
        ["ARO:3000081"],
        ["ARO:3007382"],
        ["ARO:3004022"],
        ["ARO:3000199"],
        ["ARO:3003998"],
        ["ARO:3007382", "ARO:0000052", "ARO:0000007", "ARO:0000035", "ARO:0000049"],
        ['ARO:3007145']
    ]

    expected_output = [
        ["ARO:3000007"],
        ["ARO:3000157"],
        ["ARO:3000050"],
        ["ARO:0000001"],
        ["ARO:3000081"],
        ["ARO:0000016"],
        ["ARO:0000026"],
        ["ARO:3000053"],
        ["ARO:3000007"],
        ["ARO:0000016", "ARO:0000016", "ARO:0000016", "ARO:0000016", "ARO:0000016"],
        ["ARO:3000007"]
    ]

    for t, e in zip(test_cases, expected_output):
        assert drugs_to_drug_classes(t) == e

def test_betalactams():
    test_cases = [
        _assert_aro_name('ARO:0000004', 'monobactam'),
        _assert_aro_name('ARO:0000032', 'cephalosporin'),
        _assert_aro_name('ARO:3007145', 'imipenem-cilastatin-relebactam'),
        _assert_aro_name('ARO:3003998', 'ampicillin-sulbactam'),
        _assert_aro_name('ARO:3004705', 'ceftazidime-clavulanic acid'),
        _assert_aro_name('ARO:3007040', 'cefotaxime-ceftiofur-tazobactam-clavulanate')
    ]

    beta_lactam_antibiotic = _assert_aro_name('ARO:3000007', 'beta-lactam antibiotic')
    assert set(drugs_to_drug_classes(test_cases)) == set([beta_lactam_antibiotic])

def test_peptide_antibiotics():
    test_cases = [
        _assert_aro_name('ARO:3000123', 'gramicidin'),
        _assert_aro_name('ARO:3000119', 'edeine'),
        _assert_aro_name('ARO:0000041', 'bacitracin'),
        _assert_aro_name('ARO:3000199', 'gramicidin D')
    ]

    peptide_antibiotic = _assert_aro_name('ARO:3000053', 'peptide antibiotic')
    assert set(drugs_to_drug_classes(test_cases)) == set([peptide_antibiotic])

def test_aminoglycosides():
    test_cases = [
        _assert_aro_name('ARO:3007382', 'gentamicin'),
        _assert_aro_name('ARO:0000052', 'tobramycin'),
        _assert_aro_name('ARO:0000007', 'dibekacin'),
        _assert_aro_name('ARO:0000035', 'sisomicin'),
        _assert_aro_name('ARO:0000049', 'kanamycin A')
    ]

    aminoglycoside_antibiotic = _assert_aro_name('ARO:0000016', 'aminoglycoside antibiotic')
    assert set(drugs_to_drug_classes(test_cases)) == set([aminoglycoside_antibiotic])

def test_streptogramin_mixture():
    quinupristin_dalfopristin = _assert_aro_name('ARO:3004022', 'quinupristin-dalfopristin')

    streptogramin_antibiotic = _assert_aro_name('ARO:0000026', 'streptogramin antibiotic')
    assert drugs_to_drug_classes([quinupristin_dalfopristin]) == [streptogramin_antibiotic]


def test_multiple_parents():
    # ARO:3000130 has multiple parents
    edeine_A = _assert_aro_name('ARO:3000130', 'edeine A')
    assert drugs_to_drug_classes([edeine_A]) == [_assert_aro_name('ARO:3000053', 'peptide antibiotic')]


def test_categories_all():
    import pandas as pd
    golden = pd.read_csv('tests/all_aros_category.tsv.gz',
                         sep='\t',
                         index_col=0)
    golden = golden.T.to_dict(orient='list')

    ARO = get_aro_ontology()

    for ar in ARO.terms():
        ar = ar.id
        drugs = list(confers_resistance_to(ar))
        # the walrus operator is only available in python 3.8+, so we have to
        # do this in two steps
        expected = golden.get(ar)
        if expected:
            cats = drugs_to_drug_classes(drugs)
            drugs = ';'.join(drugs)
            cats = ';'.join(cats)
            assert [drugs, cats] == expected
        else:
            assert not drugs

